package edu.kit.iti.formal.lights;

import org.sat4j.specs.IVecInt;

import java.util.LinkedList;
import java.util.List;

/**
 * A checker for validity of {@link Solution} of the LightsUp puzzles.
 * <p>
 * A solution is valid if every non-wall field is lit, and every wall constraint is met.
 * <p>
 * The class build up a light field given the lights in the solution by raytracing the shined area.
 *
 * @author Alexander Weigl
 * @version 1 (03.09.19)
 * @see #getErrors()
 * @see #isValid()
 */
public class SolutionChecker {
    private final int dimension;
    private final Lights lights;
    private List<String> errors = new LinkedList<>();
    private boolean[] lightField;
    private IVecInt lightX;
    private IVecInt lightY;

    public SolutionChecker(Lights lights, Solution solution) {
        this.lights = lights;
        lightX = solution.lightX;
        lightY = solution.lightY;
        dimension = lights.getDimension();
        lightField = getLightField();
        checkValidity();
    }

    public boolean isValid() {
        return errors.isEmpty();
    }

    public List<String> getErrors() {
        return errors;
    }

    private void checkValidity() throws IllegalStateException {
        for (int row = 0; row < lights.getDimension(); row++) {
            for (int column = 0; column < lights.getDimension(); column++) {
                if (lights.isConstrainedBlock(row, column)) {
                    var value = lights.getBlockConstraint(row, column);
                    if (isValidAndLamp(row - 1, column)) value -= 1;
                    if (isValidAndLamp(row + 1, column)) value -= 1;
                    if (isValidAndLamp(row, column + 1)) value -= 1;
                    if (isValidAndLamp(row, column - 1)) value -= 1;
                    if (value > 0) {
                        addError("Not enough lamps in neighbourhood of for constraint block: (%d,%d).", row, column);
                    }
                } else if (lights.isBlock(row, column)) {
                    continue;
                } else if (!isValidAndLit(row, column)) {
                    addError("Normal cell (%d,%d) is not lit.", row, column);
                }
            }
        }
    }

    boolean isValidAndLit(int row, int column) {
        return isValid(row, column) && lightField[pos(row, column)];
    }

    private boolean isValid(int row, int column) {
        return 0 <= row && 0 <= column && column < lights.getDimension() && row < lights.getDimension();
    }

    boolean isValidAndLamp(int row, int column) {
        if (!isValid(row, column)) return false;
        for (int i = 0; i < lightX.size(); i++) {
            int r = lightX.get(i);
            int c = lightY.get(i);
            if (r == row && c == column) {
                return true;
            }
        }
        return false;
    }

    private boolean[] getLightField() {
        lightField = new boolean[lights.getDimension() * lights.getDimension()];
        for (int i = 0; i < lightX.size(); i++) {
            int x = lightX.get(i);
            int y = lightY.get(i);
            fillLightField(x, y);
        }
        return lightField;
    }

    private void fillLightField(int row, int column) {
        lightField[pos(row, column)] = true;

        for (int r = row - 1; r >= 0; r--) {
            if (incrOrBreak(r, column)) break;
        }

        for (int r = row + 1; r < lights.getDimension(); r++) {
            if (incrOrBreak(r, column)) break;
        }

        for (int c = column - 1; c >= 0; c--) {
            if (incrOrBreak(row, c)) break;
        }

        for (int c = column + 1; c < lights.getDimension(); c++) {
            if (incrOrBreak(row, c)) break;
        }
    }

    private boolean incrOrBreak(int row, int column) {
        if (isValidAndLamp(row, column)) {
            addError("Collision with lamp at (%d,%d).", row, column);
        }

        if (lights.isBlock(row, column)) {
            return true;
        } else {
            lightField[pos(row, column)] = true;
        }
        return false;
    }

    private void addError(String message, Object... args) {
        errors.add(String.format(message, args));
    }

    private int pos(int row, int column) {
        return row * dimension + column;
    }
}
